@@ -26,9 +26,15 @@
 int virCHProcessStart(virCHDriver *driver,
                       virDomainObj *vm,
                       virDomainRunningReason reason);
+
+typedef enum {
+    VIR_CH_PROCESS_STOP_FORCE = 1 << 0,
+} virCHProcessStopFlags;
+
 int virCHProcessStop(virCHDriver *driver,
                      virDomainObj *vm,
-                     virDomainShutoffReason reason);
+                     virDomainShutoffReason reason,
+                     unsigned int flags);
 
 int virCHProcessStartRestore(virCHDriver *driver,
                          virDomainObj *vm,
